#!/usr/bin/env python
# Copyright The PyTorch Lightning team.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os
import re
from typing import List

_PROJECT_ROOT = os.path.dirname(os.path.dirname(__file__))


def _load_requirements(path_dir: str, file_name: str = "requirements.txt", comment_char: str = "#") -> List[str]:
    """Load requirements from a file.

    >>> _load_requirements(_PROJECT_ROOT)  # doctest: +ELLIPSIS +NORMALIZE_WHITESPACE
    [...]
    """
    with open(os.path.join(path_dir, file_name)) as file:
        lines = [ln.strip() for ln in file.readlines()]
    reqs = []
    for ln in lines:
        # filer all comments
        comment = ""
        if comment_char in ln:
            start = ln.index(comment_char)
            comment = ln[start:]
            ln = ln[:start]
        req = ln.strip()
        # skip directly installed dependencies
        if not req or req.startswith("http") or req.startswith("http") or req.startswith("git") or "@http" in req:
            continue

        # adding strict back to the comment
        if "strict" in comment:
            req += "  # strict"

        reqs.append(req)
    return reqs


def _load_readme_description(path_dir: str, homepage: str, ver: str) -> str:
    """Load readme as decribtion."""
    path_readme = os.path.join(path_dir, "README.md")
    with open(path_readme, encoding="utf-8") as fp:
        text = fp.read()

    # https://github.com/PyTorchLightning/pytorch-lightning/raw/master/docs/source/_images/lightning_module/pt_to_pl.png
    github_source_url = os.path.join(homepage, "raw", ver)
    # replace relative repository path to absolute link to the release
    #  do not replace all "docs" as in the readme we reger some other sources with particular path to docs
    text = text.replace("docs/source/_static/", f"{os.path.join(github_source_url, 'docs/source/_static/')}")

    # readthedocs badge
    text = text.replace("badge/?version=stable", f"badge/?version={ver}")
    text = text.replace("lightning.readthedocs.io/en/stable/", f"lightning.readthedocs.io/en/{ver}")
    # codecov badge
    text = text.replace("/branch/master/graph/badge.svg", f"/release/{ver}/graph/badge.svg")
    # replace github badges for release ones
    text = text.replace("badge.svg?branch=master&event=push", f"badge.svg?tag={ver}")

    skip_begin = r"<!-- following section will be skipped from PyPI description -->"
    skip_end = r"<!-- end skipping PyPI description -->"
    # todo: wrap content as commented description
    text = re.sub(rf"{skip_begin}.+?{skip_end}", "<!--  -->", text, flags=re.IGNORECASE + re.DOTALL)

    # # https://github.com/Borda/pytorch-lightning/releases/download/1.1.0a6/codecov_badge.png
    # github_release_url = os.path.join(homepage, "releases", "download", ver)
    # # download badge and replace url with local file
    # text = _parse_for_badge(text, github_release_url)
    return text



from typing import Any, Dict

from lightning import LightningFlow

from lightning_hpo.loggers.logger import Logger
from lightning_hpo.loggers.streamlit.hyperplot import HiPlotFlow


class StreamLitLogger(Logger):
    def __init__(self):
        super().__init__()
        self.hi_plot = HiPlotFlow()

    def on_after_experiment_end(
        self, sweep_id: str, experiment_id: int, monitor: str, score: float, params: Dict[str, Any]
    ):
        self.hi_plot.data.append({monitor: score, **params})

    def connect(self, flow: LightningFlow):
        flow.hi_plot = self.hi_plot

    def configure_layout(self):
        return [{"name": "Experiment", "content": self.hi_plot}]

    def configure_tracer(self, tracer, sweep_id: str, experiment_id: int, experiment_name: str, params: Dict[str, Any]):
        pass

    def get_url(self, experiment_id: int) -> None:
        return None



import functools
import json
from dataclasses import dataclass
from typing import Generic, Optional, Type, TypeVar, Union

from fastapi.encoders import jsonable_encoder
from lightning import CloudCompute as LightningCloudCompute
from lightning import LightningFlow
from lightning.app.storage import Path
from lightning.app.utilities.enum import WorkStageStatus
from pydantic import parse_obj_as
from pydantic.main import ModelMetaclass
from sqlalchemy.inspection import inspect
from sqlmodel import JSON, SQLModel, TypeDecorator

from lightning_hpo.framework import _OBJECTIVE_FRAMEWORK
from lightning_hpo.framework.agnostic import Objective

T = TypeVar("T")


@dataclass
class HPOCloudCompute(LightningCloudCompute):
    count: int = 1


def get_best_model_score(flow: LightningFlow) -> Optional[float]:
    metrics = [work.best_model_score for work in flow.works()]
    if not metrics or not all(metrics):
        return None
    return max(metrics)


def get_best_model_path(flow: LightningFlow) -> Optional[Path]:
    metrics = {work.best_model_score: work for work in flow.works()}
    if not all(metrics) or all(metric is None for metric in metrics):
        return None

    return metrics[max(metrics)].best_model_path


def _resolve_objective_cls(objective_cls, framework: str):
    if objective_cls is None:
        if framework not in _OBJECTIVE_FRAMEWORK:
            raise Exception(f"The supported framework are {list(_OBJECTIVE_FRAMEWORK)}. Found {framework}.")
        objective_cls = _OBJECTIVE_FRAMEWORK[framework]

    return objective_cls


def _check_stage(obj: Union[LightningFlow, Objective], status: str) -> bool:
    if isinstance(obj, Objective):
        return obj.status.stage == status
    else:
        works = obj.works()
        if works:
            return any(w.status.stage == status for w in obj.works())
        else:
            return status == WorkStageStatus.NOT_STARTED


# Taken from https://github.com/tiangolo/sqlmodel/issues/63#issuecomment-1081555082
def pydantic_column_type(pydantic_type):
    class PydanticJSONType(TypeDecorator, Generic[T]):
        impl = JSON()

        def __init__(
            self,
            json_encoder=json,
        ):
            self.json_encoder = json_encoder
            super(PydanticJSONType, self).__init__()

        def bind_processor(self, dialect):
            impl_processor = self.impl.bind_processor(dialect)
            dumps = self.json_encoder.dumps
            if impl_processor:

                def process(value: T):
                    if value is not None:
                        if isinstance(pydantic_type, ModelMetaclass):
                            # This allows to assign non-InDB models and if they're
                            # compatible, they're directly parsed into the InDB
                            # representation, thus hiding the implementation in the
                            # background. However, the InDB model will still be returned
                            value_to_dump = pydantic_type.from_orm(value)
                        else:
                            value_to_dump = value
                        value = jsonable_encoder(value_to_dump)
                    return impl_processor(value)

            else:

                def process(value):
                    if isinstance(pydantic_type, ModelMetaclass):
                        # This allows to assign non-InDB models and if they're
                        # compatible, they're directly parsed into the InDB
                        # representation, thus hiding the implementation in the
                        # background. However, the InDB model will still be returned
                        value_to_dump = pydantic_type.from_orm(value)
                    else:
                        value_to_dump = value
                    value = dumps(jsonable_encoder(value_to_dump))
                    return value

            return process

        def result_processor(self, dialect, coltype) -> T:
            impl_processor = self.impl.result_processor(dialect, coltype)
            if impl_processor:

                def process(value):
                    value = impl_processor(value)
                    if value is None:
                        return None

                    data = value
                    # Explicitly use the generic directly, not type(T)
                    full_obj = parse_obj_as(pydantic_type, data)
                    return full_obj

            else:

                def process(value):
                    if value is None:
                        return None

                    # Explicitly use the generic directly, not type(T)
                    full_obj = parse_obj_as(pydantic_type, value)
                    return full_obj

            return process

        def compare_values(self, x, y):
            return x == y

    return PydanticJSONType


@functools.lru_cache
def get_primary_key(model_type: Type[SQLModel]) -> str:
    primary_keys = inspect(model_type).primary_key

    if len(primary_keys) != 1:
        raise ValueError(f"The model {model_type.__name__} should have a single primary key field.")

    return primary_keys[0].name



import sys
from typing import List

from lightning.app.utilities.commands import ClientCommand
from rich.console import Console
from rich.table import Table

from lightning_hpo.commands.sweep.run import SweepConfig


def _show_experiments(sweeps: List[SweepConfig]):
    table = Table(
        "name",
        "cloud compute",
        "progress",
        "best model score",
        "sweep name",
        title="Experiments",
        show_header=True,
        header_style="bold green",
    )

    for sweep in sweeps:
        experiments = sweep.experiments.values()
        for experiment in experiments:
            table.add_row(
                str(experiment.name),
                sweep.cloud_compute,
                str(experiment.progress) if experiment.stage != "failed" else "failed",
                str(round(experiment.best_model_score, 2) if experiment.best_model_score else None),
                None if len(experiments) == 1 else sweep.sweep_id,
            )
    console = Console()
    console.print(table)


class ShowExperimentsCommand(ClientCommand):

    description = "Show experiments and their statuses."

    def run(self) -> None:
        if sys.argv and sys.argv[-1] == "--help":
            print("optional arguments:")
            print("  -h, --help   show this help message and exit")
            return

        # 1: Collect the SweepConfig
        resp = self.invoke_handler()

        # 3: Display the Sweeps or Sweep
        # TODO: Undestand why the format isn't the same
        try:
            sweeps = [SweepConfig.parse_raw(sweep) for sweep in resp]
        except Exception:
            sweeps = [SweepConfig(**sweep) for sweep in resp]

        _show_experiments(sweeps)



import os
from uuid import uuid4

from lightning import CloudCompute, LightningFlow
from lightning.app.components.database import Database, DatabaseClient
from lightning.app.frontend import StaticWebFrontend
from lightning.app.storage import Drive

from lightning_hpo.commands.artifacts.download import (
    _collect_artifact_urls,
    DownloadArtifactsCommand,
    DownloadArtifactsConfig,
    DownloadArtifactsConfigResponse,
)
from lightning_hpo.commands.artifacts.show import (
    _collect_artifact_paths,
    ShowArtifactsCommand,
    ShowArtifactsConfig,
    ShowArtifactsConfigResponse,
)
from lightning_hpo.commands.data.create import CreateDataCommand, DataConfig
from lightning_hpo.commands.data.delete import DeleteDataCommand, DeleteDataConfig
from lightning_hpo.commands.data.show import ShowDataCommand
from lightning_hpo.commands.logs.show import ShowLogsCommand

# from lightning_hpo.controllers.notebook import NotebookController
from lightning_hpo.controllers.sweep import SweepController
from lightning_hpo.controllers.tensorboard import TensorboardController


class TrainingStudio(LightningFlow):
    def __init__(self):
        super().__init__()
        # 1: Create Drive
        self.drive = Drive("lit://uploaded_files")

        # 2: Controllers
        self.sweep_controller = SweepController(self.drive)
        # self.notebook_controller = NotebookController()
        self.tensorboard_controller = TensorboardController()

        # 4: Create the database.
        self.db = Database(
            models=[
                self.sweep_controller.model,
                # self.notebook_controller.model,
                self.tensorboard_controller.model,
                DataConfig,
            ]
        )
        self.db.cloud_compute = CloudCompute("cpu", disk_size=80)

        self.has_setup = False
        self._db_client = None
        self._token = uuid4().hex
        self.seconds = ",".join([str(v) for v in range(0, 60, 1)])

    @property
    def ready(self) -> bool:
        return self.db.alive()

    def run(self):
        self.db.run(token=self._token)

        if not self.db.alive():
            return

        if not self.has_setup:
            print("The Training Studio App is ready !")
            self.has_setup = True

        # Run every seconds.
        if self.schedule(f"* * * * * {self.seconds}"):

            # 3: Run the controllers
            self.sweep_controller.run(self.db.db_url, token=self._token)
            # self.notebook_controller.run(self.db.db_url)
            self.tensorboard_controller.run(self.db.db_url, token=self._token)

    def configure_layout(self):
        return StaticWebFrontend(os.path.join(os.path.dirname(__file__), "ui", "build"))

    def show_artifacts(self, config: ShowArtifactsConfig) -> ShowArtifactsConfigResponse:
        sweeps = self.db_client.select_all(self.sweep_controller.model)
        return ShowArtifactsConfigResponse(
            sweep_names=[sweep.sweep_id for sweep in sweeps],
            experiment_names=[exp.name for sweep in sweeps for exp in sweep.experiments.values()],
            paths=_collect_artifact_paths(config),
        )

    def download_artifacts(self, config: DownloadArtifactsConfig) -> DownloadArtifactsConfigResponse:
        sweeps = self.db_client.select_all(self.sweep_controller.model)
        paths, urls = _collect_artifact_urls(config)
        return DownloadArtifactsConfigResponse(
            sweep_names=[sweep.sweep_id for sweep in sweeps],
            experiment_names=[exp.name for sweep in sweeps for exp in sweep.experiments.values()],
            paths=paths,
            urls=urls,
        )

    def create_data(self, config: DataConfig):
        drives = self.db_client.select_all(DataConfig)
        for drive in drives:
            if drive.name == config.name:
                return f"The data `{config.name}` already exists."
        self.db_client.insert(config)
        return f"The data `{config.name}` has been created."

    def delete_data(self, config: DeleteDataConfig):
        drives = self.db_client.select_all(DataConfig)
        for drive in drives:
            if drive.name == config.name:
                self.db_client.delete(drive)
                return f"The data `{config.name}` has been deleted."
        return f"The data `{config.name}` doesn't exist."

    def show_data(self):
        if self.db.db_url:
            return self.db_client.select_all(DataConfig)
        return []

    def configure_commands(self):
        controller_commands = self.sweep_controller.configure_commands()
        # controller_commands += self.notebook_controller.configure_commands()
        controller_commands += [
            {"show artifacts": ShowArtifactsCommand(self.show_artifacts)},
            {"download artifacts": DownloadArtifactsCommand(self.download_artifacts)},
            {"create data": CreateDataCommand(self.create_data)},
            {"delete data": DeleteDataCommand(self.delete_data)},
            {"show data": ShowDataCommand(self.show_data)},
            {"show logs": ShowLogsCommand(self.sweep_controller.show_logs)},
        ]
        return controller_commands

    def configure_api(self):
        return self.tensorboard_controller.configure_api()

    @property
    def db_client(self) -> DatabaseClient:
        if self._db_client is None:
            assert self.db.db_url is not None
            self._db_client = DatabaseClient(self.db.db_url, token=self._token)
        return self._db_client



from argparse import ArgumentParser

from lightning.app.utilities.commands import ClientCommand
from pydantic import BaseModel


class StopNotebookConfig(BaseModel):
    notebook_name: str


class StopNotebookCommand(ClientCommand):
    def run(self) -> None:
        parser = ArgumentParser()
        parser.add_argument("--name", required=True, help="The name of your notebook to be stopped.")
        hparams, _ = parser.parse_known_args()
        response = self.invoke_handler(config=StopNotebookConfig(notebook_name=hparams.name))
        print(response)



import os.path as ops

import optuna
from lightning import LightningApp

from lightning_hpo import Sweep
from lightning_hpo.algorithm.optuna import OptunaAlgorithm
from lightning_hpo.distributions.distributions import Categorical, IntUniform, LogUniform, Uniform

app = LightningApp(
    Sweep(
        script_path=ops.join(ops.dirname(__file__), "scripts/train.py"),
        total_experiments=3,
        parallel_experiments=1,
        distributions={
            "model.lr": LogUniform(0.001, 0.1),
            "model.gamma": Uniform(0.5, 0.8),
            "data.batch_size": Categorical([16, 32, 64]),
            "trainer.max_epochs": IntUniform(1, 5),
        },
        algorithm=OptunaAlgorithm(optuna.create_study(direction="maximize")),
        framework="pytorch_lightning",
    )
)



import os
import uuid
from typing import Any, Dict, List, Optional, Tuple, Type, Union
from uuid import uuid4

from lightning import BuildConfig, CloudCompute, LightningFlow
from lightning.app.components.python.tracer import Code
from lightning.app.frontend import StaticWebFrontend
from lightning.app.storage.mount import Mount
from lightning.app.storage.path import Path
from lightning.app.utilities.app_helpers import _LightningAppRef
from lightning_utilities.core.apply_func import apply_to_collection

from lightning_hpo.algorithm.base import Algorithm
from lightning_hpo.algorithm.optuna import GridSearch, OptunaAlgorithm, RandomSearch
from lightning_hpo.commands.sweep.run import ExperimentConfig, SweepConfig
from lightning_hpo.commands.sweep.show import ShowSweepsCommand
from lightning_hpo.controllers.controller import ControllerResource
from lightning_hpo.distributions.distributions import Distribution
from lightning_hpo.framework.agnostic import Objective
from lightning_hpo.loggers import LoggerType
from lightning_hpo.utilities.enum import Stage
from lightning_hpo.utilities.utils import (
    _check_stage,
    _resolve_objective_cls,
    get_best_model_path,
    get_best_model_score,
    HPOCloudCompute,
)


class CustomBuildConfig(BuildConfig):
    def __init__(self, *args, packages, **kwargs):
        super().__init__(*args, **kwargs)
        self.packages = packages

    def build_commands(self):
        package_installs = [f"apt install {package}" for package in self.packages]
        return super().build_commands() + package_installs


class Sweep(LightningFlow, ControllerResource):

    model = SweepConfig

    def __init__(
        self,
        total_experiments: int,
        objective_cls: Optional[Type[Objective]] = None,
        parallel_experiments: int = 1,
        script_args: Optional[Union[list, str]] = None,
        env: Optional[Dict] = None,
        cloud_compute: Optional[HPOCloudCompute] = None,
        script_path: Optional[str] = None,
        algorithm: Optional[Algorithm] = None,
        logger: Optional[str] = "streamlit",
        sweep_id: Optional[str] = None,
        distributions: Optional[Dict[str, Union[Dict, Distribution]]] = None,
        framework: str = "base",
        code: Optional[Code] = None,
        direction: Optional[str] = None,
        total_experiments_done: Optional[int] = 0,
        requirements: Optional[List[str]] = None,
        packages: Optional[List[str]] = None,
        experiments: Optional[Dict[int, Dict]] = None,
        stage: Optional[str] = Stage.NOT_STARTED,
        logger_url: str = "",
        pip_install_source: bool = False,
        data: Optional[List[Tuple[str, str]]] = None,
        **objective_kwargs: Any,
    ):
        """The Sweep class enables to easily run a Python Script with Lightning
        :class:`~lightning.utilities.tracer.Tracer` with state-of-the-art distributed.
        Arguments:
            total_experiments: Number of HPO experiments to run.
            objective_cls: Your custom base objective work.
            parallel_experiments: Number of parallel experiments to run.
            script_args: Optional script arguments.
            env: Environment variables to be passed to the script.
            cloud_compute: The cloud compute on which the Work should run on.
            blocking: Whether the Work should be blocking or asynchronous.
            script_path: Path of the python script to run.
            logger: Which logger to use
            objective_kwargs: Your custom keywords arguments passed to your custom objective work class.
        """
        super().__init__()
        # Serialize the distributions
        distributions = apply_to_collection(distributions, Distribution, lambda x: x.to_dict())

        # SweepConfig
        self.sweep_id = sweep_id or str(uuid.uuid4()).split("-")[0]
        self.script_path = script_path
        self.total_experiments = total_experiments
        self.parallel_experiments = parallel_experiments
        self.total_experiments_done = total_experiments_done or 0
        self.requirements = requirements or []
        self.packages = packages or []
        self.script_args = script_args
        self.distributions = distributions or {}
        self.framework = framework
        self.cloud_compute = getattr(cloud_compute, "name", "default")
        self.num_nodes = getattr(cloud_compute, "count", 1) if cloud_compute else 1
        self.disk_size = getattr(cloud_compute, "disk_size", 1) if cloud_compute else 10
        self.logger = logger
        self.direction = direction
        self.experiments = experiments or {}
        self.stage = stage
        self.logger_url = logger_url
        self.pip_install_source = pip_install_source
        self.data = data

        self._objective_cls = _resolve_objective_cls(objective_cls, framework)
        self._algorithm = algorithm or OptunaAlgorithm(direction=direction)
        self._logger = LoggerType(logger).get_logger()
        self._logger.connect(self)

        self._kwargs = {
            "script_path": script_path,
            "env": env,
            "script_args": script_args,
            "num_nodes": getattr(cloud_compute, "count", 1) if cloud_compute else 1,
            "logger": logger,
            "code": code,
            "sweep_id": self.sweep_id,
            "raise_exception": False,
            "cloud_build_config": CustomBuildConfig(requirements=self.requirements, packages=self.packages),
            **objective_kwargs,
        }
        self._algorithm.register_distributions(self.distributions)
        self._algorithm.register_experiments(
            [t for t in experiments.values() if t["stage"] == Stage.SUCCEEDED] if experiments else []
        )
        self.restart_count = 0

    def run(self):
        if self.stage in (Stage.SUCCEEDED, Stage.STOPPED):
            return

        if self.total_experiments_done == self.total_experiments:
            self.stage = Stage.SUCCEEDED
            return

        for experiment_id in range(self.num_experiments):

            objective = self._get_objective(experiment_id)

            if objective:

                if _check_stage(objective, Stage.NOT_STARTED):
                    self._algorithm.experiment_start(experiment_id)
                    self._logger.on_after_experiment_start(self.sweep_id)

                if not self.experiments[experiment_id]["params"]:
                    self.stage = Stage.RUNNING
                    self.experiments[experiment_id]["params"] = self._algorithm.get_params(experiment_id)

                logger_url = self._logger.get_url(experiment_id)
                if logger_url is not None and self.logger_url != logger_url:
                    self.logger_url = logger_url

                if _check_stage(objective, Stage.FAILED):
                    continue

                objective.run(
                    params=self._algorithm.get_params(experiment_id),
                    restart_count=self.restart_count,
                )

                self.experiments[experiment_id]["progress"] = objective.progress
                self.experiments[experiment_id]["total_parameters"] = getattr(objective, "total_parameters", None)
                self.experiments[experiment_id]["start_time"] = getattr(objective, "start_time", None)
                self.experiments[experiment_id]["end_time"] = getattr(objective, "end_time", None)
                self.experiments[experiment_id]["best_model_score"] = getattr(objective, "best_model_score", None)
                self.experiments[experiment_id]["last_model_path"] = str(getattr(objective, "last_model_path", ""))

                if _check_stage(objective, Stage.FAILED):
                    self.experiments[experiment_id]["stage"] = Stage.FAILED
                    self.experiments[experiment_id]["exception"] = objective.status.message
                    objective.stop()

                if objective.reports and not self.experiments[experiment_id]["stage"] == "pruned":
                    if self._algorithm.should_prune(experiment_id, objective.reports):
                        self.experiments[experiment_id]["stage"] = Stage.PRUNED
                        objective.stop()
                        self.total_experiments_done += 1
                        continue

                if self.stage != Stage.FAILED and self.experiments[experiment_id]["stage"] == Stage.PENDING:
                    if self.experiments[experiment_id]["stage"] != objective.status:
                        self.stage = Stage.RUNNING
                        self.experiments[experiment_id]["stage"] = Stage.RUNNING

                if objective.best_model_score:
                    if self.experiments[experiment_id]["stage"] == Stage.SUCCEEDED:
                        pass
                    elif self.experiments[experiment_id]["stage"] not in (Stage.PRUNED, Stage.STOPPED, Stage.FAILED):
                        self._algorithm.experiment_end(experiment_id, objective.best_model_score)
                        self._logger.on_after_experiment_end(
                            sweep_id=self.sweep_id,
                            experiment_id=objective.experiment_id,
                            monitor=objective.monitor,
                            score=objective.best_model_score,
                            params=self._algorithm.get_params(experiment_id),
                        )
                        self.experiments[experiment_id]["best_model_score"] = objective.best_model_score
                        self.experiments[experiment_id]["best_model_path"] = str(objective.best_model_path)
                        self.experiments[experiment_id]["monitor"] = objective.monitor
                        self.experiments[experiment_id]["stage"] = Stage.SUCCEEDED
                        self.total_experiments_done += 1
                        objective.stop()

        if all(
            self.experiments[experiment_id]["stage"] == Stage.FAILED for experiment_id in range(self.num_experiments)
        ):
            self.stage = Stage.FAILED

    @property
    def num_experiments(self) -> int:
        return min(self.total_experiments_done + self.parallel_experiments, self.total_experiments)

    @property
    def best_model_score(self) -> Optional[float]:
        return get_best_model_score(self)

    @property
    def best_model_path(self) -> Optional[Path]:
        return get_best_model_path(self)

    def stop_experiment(self, experiment_id: int):
        objective = self._get_objective(experiment_id)
        if objective:
            objective.stop()
            self.experiments[experiment_id]["stage"] = Stage.STOPPED
            self.total_experiments_done += 1

    def _get_objective(self, experiment_id: int):
        experiment_config = self.experiments.get(experiment_id, None)
        if experiment_config is None:
            experiment_config = ExperimentConfig(
                name=str(uuid4()).split("-")[-1][:7],
                best_model_score=None,
                monitor=None,
                best_model_path=None,
                stage=Stage.PENDING,
                params={},
            ).dict()
            self.experiments[experiment_id] = experiment_config

        if experiment_config["stage"] == Stage.SUCCEEDED:
            return

        objective = getattr(self, f"w_{experiment_id}", None)
        if objective is None:
            cloud_compute = CloudCompute(
                name=self.cloud_compute if self.cloud_compute else "cpu",
                disk_size=self.disk_size,
                mounts=[Mount(source, mount_path) for source, mount_path in self.data] if self.data else None,
            )
            objective = self._objective_cls(
                experiment_id=experiment_id,
                experiment_name=experiment_config["name"],
                cloud_compute=cloud_compute,
                last_model_path=experiment_config["last_model_path"],
                pip_install_source=self.pip_install_source,
                **self._kwargs,
            )
            setattr(self, f"w_{experiment_id}", objective)
            self.experiments[experiment_id]["stage"] = Stage.PENDING
        return objective

    @classmethod
    def from_config(
        cls, config: SweepConfig, code: Optional[Code] = None, data: Optional[List[Tuple[str, str]]] = None
    ):

        if config.algorithm == "grid_search":
            algorithm = GridSearch({k: v.dict()["params"]["choices"] for k, v in config.distributions.items()})
            config.total_experiments = algorithm.total_experiments
            config.parallel_experiments = algorithm.total_experiments

        elif config.algorithm == "random_search":
            algorithm = RandomSearch({k: v.dict() for k, v in config.distributions.items()})
        else:
            algorithm = OptunaAlgorithm(direction=config.direction)

        return cls(
            script_path=config.script_path,
            total_experiments=config.total_experiments,
            parallel_experiments=config.parallel_experiments,
            framework=config.framework,
            script_args=config.script_args,
            total_experiments_done=config.total_experiments_done,
            distributions={k: v.dict() for k, v in config.distributions.items()},
            cloud_compute=HPOCloudCompute(
                config.cloud_compute,
                count=config.num_nodes,
                disk_size=config.disk_size,
            ),
            sweep_id=config.sweep_id,
            code=code,
            logger=config.logger,
            algorithm=algorithm,
            experiments={k: v.dict() for k, v in config.experiments.items()},
            direction=config.direction,
            stage=config.stage,
            logger_url=config.logger_url,
            data=data,
            pip_install_source=config.pip_install_source,
            requirements=config.requirements,
            packages=config.packages,
        )

    def configure_layout(self):
        app = _LightningAppRef().get_current()
        if app and app.root == self:
            return StaticWebFrontend(os.path.join(os.path.dirname(os.path.dirname(__file__)), "app", "ui", "build"))
        return self._logger.configure_layout()

    def show_sweeps(self):
        return [self.collect_model()]

    def show_tensorboards(self):
        return []

    def configure_commands(self):
        return [
            {"show sweeps": ShowSweepsCommand(self.show_sweeps)},
            {"show tensorboards": self.show_tensorboards},
        ]



import os
from argparse import ArgumentParser

from lightning.app.utilities.commands import ClientCommand
from pydantic import validator
from sqlmodel import Field, SQLModel


class DataConfig(SQLModel, table=True):

    __table_args__ = {"extend_existing": True}

    name: str = Field(primary_key=True)
    source: str
    mount_path: str

    @validator("source")
    def source_validator(cls, v):
        if not v.startswith("s3://"):
            raise Exception('The `source` needs to start with "s3://"')
        elif not v.endswith("/"):
            raise Exception("The `source` needs to end with in a trailing slash (`/`)")
        return v

    @validator("mount_path")
    def mount_path_validator(cls, v, values, **kwargs):
        if not v.startswith("/"):
            raise Exception("The `mount_path` needs to start with a leading slash (`/`)")
        elif not v.endswith("/"):
            raise Exception("The `mount_path` needs to end with in a trailing slash (`/`)")
        return v


class CreateDataCommand(ClientCommand):

    description = """Create a Data association by providing a public S3 bucket and an optional mount point.
                     The contents of the bucket can be then mounted on experiments and sweeps and
                     accessed through the filesystem."""

    def run(self) -> None:
        parser = ArgumentParser()

        parser.add_argument("--name", type=str, required=True, help="The name of the Data.")
        parser.add_argument("--source", type=str, required=True, help="The S3 URL of the Data.")
        parser.add_argument(
            "--mount_path",
            type=str,
            default=None,
            help="Where the Data should be mounted in experiments and sweeps. Defaults to `/data/<name>/`.",
        )

        hparams = parser.parse_args()
        mount_path = hparams.mount_path if hparams.mount_path else os.path.join("/data", hparams.name, "")
        response = self.invoke_handler(
            config=DataConfig(name=hparams.name, source=hparams.source, mount_path=mount_path)
        )
        print(response)



import concurrent.futures
import os
from pathlib import Path
from time import time
from typing import Any, Dict, Optional

from lightning_hpo.utilities.imports import _IS_PYTORCH_LIGHTNING_AVAILABLE

if _IS_PYTORCH_LIGHTNING_AVAILABLE:
    import pytorch_lightning
    from pytorch_lightning.loggers import TensorBoardLogger
else:
    import lightning.pytorch as pytorch_lightning
    from lightning.pytorch.loggers import TensorBoardLogger

from fsspec.implementations.local import LocalFileSystem
from lightning import LightningFlow
from lightning.app.storage import Drive
from lightning.app.storage.path import _filesystem

from lightning_hpo.loggers.logger import Logger


class DriveTensorBoardLogger(TensorBoardLogger):
    def __init__(self, *args, drive: Drive, refresh_time: int = 5, **kwargs):
        super().__init__(*args, **kwargs)
        self.timestamp = None
        self.drive = drive
        self.refresh_time = refresh_time

    def log_metrics(self, metrics, step) -> None:
        super().log_metrics(metrics, step)
        if self.timestamp is None:
            self._upload_to_storage()
            self.timestamp = time()
        elif (time() - self.timestamp) > self.refresh_time:
            self._upload_to_storage()
            self.timestamp = time()

    def finalize(self, status: str) -> None:
        super().finalize(status)

    def _upload_to_storage(self):
        fs = _filesystem()
        fs.invalidate_cache()

        source_path = Path(self.log_dir).resolve()
        destination_path = self.drive._to_shared_path(self.log_dir, component_name=self.drive.component_name)

        def _copy(from_path: Path, to_path: Path) -> Optional[Exception]:

            try:
                # NOTE: S3 does not have a concept of directories, so we do not need to create one.
                if isinstance(fs, LocalFileSystem):
                    fs.makedirs(str(to_path.parent), exist_ok=True)

                fs.put(str(from_path), str(to_path), recursive=False)

                # Don't delete tensorboard logs.
                if "events.out.tfevents" not in str(from_path):
                    os.remove(str(from_path))

            except Exception as e:
                # Return the exception so that it can be handled in the main thread
                return e

        src = [file for file in source_path.rglob("*") if file.is_file()]
        dst = [destination_path / file.relative_to(source_path) for file in src]

        with concurrent.futures.ThreadPoolExecutor(4) as executor:
            results = executor.map(_copy, src, dst)

        # Raise the first exception found
        exception = next((e for e in results if isinstance(e, Exception)), None)
        if exception:
            raise exception


class TensorboardLogger(Logger):
    def on_after_experiment_start(self, sweep_id: str, title: Optional[str] = None, desc: Optional[str] = None):
        pass

    def on_after_experiment_end(
        self, sweep_id: str, experiment_id: int, monitor: str, score: float, params: Dict[str, Any]
    ):
        pass

    def connect(self, flow: LightningFlow):
        pass

    def configure_layout(self):
        return []

    def configure_tracer(self, tracer, sweep_id: str, experiment_id: int, experiment_name: str, params: Dict[str, Any]):
        # Create a space logs under the sweep_id folder
        drive = Drive(f"lit://{sweep_id}", component_name=experiment_name, allow_duplicates=True)
        use_localhost = "LIGHTNING_APP_STATE_URL" not in os.environ

        if use_localhost:
            logger = TensorBoardLogger(save_dir=str(drive.root), name="", version="")
        else:
            logger = DriveTensorBoardLogger(save_dir=".", name="", drive=drive, refresh_time=5)

        # TODO: Collect the monitor + metric at the end.
        logger.log_hyperparams(params)

        def trainer_pre_fn(trainer, *args, **kwargs):
            kwargs["logger"] = logger
            return {}, args, kwargs

        tracer.add_traced(pytorch_lightning.Trainer, "__init__", pre_fn=trainer_pre_fn)

    def get_url(self, experiment_id: int) -> None:
        pass



import os
import re
from argparse import ArgumentParser
from ast import literal_eval
from getpass import getuser
from pathlib import Path
from typing import Dict, List, Optional, Union
from uuid import uuid4

import requests
from hydra.core.override_parser.overrides_parser import OverridesParser
from lightning.app.source_code import LocalSourceCodeDir
from lightning.app.source_code.uploader import FileUploader
from lightning.app.utilities.commands import ClientCommand
from pydantic import validator
from sqlalchemy import Column
from sqlmodel import Field, SQLModel

from lightning_hpo.loggers import LoggerType
from lightning_hpo.utilities.enum import Stage
from lightning_hpo.utilities.utils import pydantic_column_type

RANGE_REGEX: re.Pattern = re.compile(r"^range\(([0-9]+,( ){0,1}){0,2}[0-9]+\)$")


class Distributions(SQLModel, table=False):
    distribution: str
    params: Dict[str, Union[float, int, str, List[float], List[str]]] = Field(
        sa_column=Column(pydantic_column_type(Dict[str, Union[float, int, List[float], List[str]]]))
    )


class ExperimentConfig(SQLModel, table=False):
    name: str
    best_model_score: Optional[float]
    monitor: Optional[str]
    best_model_path: Optional[str]
    last_model_path: Optional[str]
    stage: str = Stage.NOT_STARTED
    params: Dict[str, Union[float, int, str, List[float], List[str]]] = Field(
        sa_column=Column(pydantic_column_type(Dict[str, Union[float, int, List[float], List[str]]]))
    )
    exception: Optional[str]
    progress: Optional[float]
    total_parameters: Optional[str] = None
    start_time: Optional[str] = None
    end_time: Optional[str] = None

    @property
    def pruned(self) -> bool:
        return self.stage == Stage.PRUNED


class SweepConfig(SQLModel, table=True):
    __table_args__ = {"extend_existing": True}

    sweep_id: str = Field(primary_key=True)
    script_path: str
    total_experiments: int
    parallel_experiments: int
    total_experiments_done: int = 0
    requirements: List[str] = Field(..., sa_column=Column(pydantic_column_type(List[str])))
    packages: List[str] = Field(..., sa_column=Column(pydantic_column_type(List[str])))
    script_args: List[str] = Field(..., sa_column=Column(pydantic_column_type(List[str])))
    algorithm: str
    distributions: Dict[str, Distributions] = Field(
        ..., sa_column=Column(pydantic_column_type(Dict[str, Distributions]))
    )
    logger_url: str = ""
    experiments: Dict[int, ExperimentConfig] = Field(
        ..., sa_column=Column(pydantic_column_type(Dict[int, ExperimentConfig]))
    )
    framework: str
    cloud_compute: str = "cpu"
    num_nodes: int = 1
    logger: str
    direction: str
    stage: str = Stage.NOT_STARTED
    desired_stage: str = Stage.RUNNING
    disk_size: int = 80
    pip_install_source: bool = False
    data: Dict[str, Optional[str]] = Field(..., sa_column=Column(pydantic_column_type(Dict[str, Optional[str]])))
    username: Optional[str] = None

    @property
    def num_experiments(self) -> int:
        return min(self.total_experiments_done + self.parallel_experiments, self.total_experiments)

    @property
    def hash(self) -> str:
        return self.sweep_id.split("-")[1]

    def is_tensorboard(self):
        return self.logger == LoggerType.TENSORBOARD.value

    @validator("data")
    def data_validator(cls, v, values, **kwargs):
        for _, mount_path in v.items():
            if mount_path is not None:
                if not mount_path.startswith("/"):
                    raise Exception("The `mount_path` needs to start with a leading slash (`/`)")
                elif not mount_path.endswith("/"):
                    raise Exception("The `mount_path` needs to end with in a trailing slash (`/`)")
        return v


class DistributionParser:
    @staticmethod
    def is_distribution(argument: str) -> bool:
        ...

    @staticmethod
    def parse(argument: str) -> Dict:
        ...


class UniformDistributionParser(DistributionParser):
    @staticmethod
    def is_distribution(argument: str) -> bool:
        return "uniform" in argument

    @staticmethod
    def parse(argument: str):
        name, value = argument.split("=")
        regex = "[0-9]*\.[0-9]*"  # noqa W605
        low, high = re.findall(regex, value)
        return {name: {"distribution": "uniform", "params": {"low": float(low), "high": float(high)}}}


class LogUniformDistributionParser(DistributionParser):
    @staticmethod
    def is_distribution(argument: str) -> bool:
        return "log_uniform" in argument

    @staticmethod
    def parse(argument: str):
        name, value = argument.split("=")
        regex = "[0-9]+\.[0-9]+|[0-9]+"  # noqa W605
        low, high = re.findall(regex, value)
        return {name: {"distribution": "log_uniform", "params": {"low": float(low), "high": float(high)}}}


class CategoricalDistributionParser(DistributionParser):
    @staticmethod
    def is_distribution(argument: str) -> bool:
        return "categorical" in argument

    @staticmethod
    def parse(argument: str):
        name, value = argument.split("=")
        choices = value.split("[")[1].split("]")[0].split(", ")
        return {name: {"distribution": "categorical", "params": {"choices": choices}}}


class CustomFileUploader(FileUploader):
    def _upload_data(self, s: requests.Session, url: str, data: bytes):
        resp = s.put(url, files={"uploaded_file": data})
        assert resp.status_code == 200


class CustomLocalSourceCodeDir(LocalSourceCodeDir):
    def upload(self, url: str) -> None:
        """Uploads package to URL, usually pre-signed URL.

        Notes
        -----
        Since we do not use multipart uploads here, we cannot upload any
        packaged repository files which have a size > 2GB.

        This limitation should be removed during the datastore upload redesign
        """
        if self.package_path.stat().st_size > 2e9:
            raise OSError(
                "cannot upload directory code whose total fize size is greater than 2GB (2e9 bytes)"
            ) from None

        uploader = CustomFileUploader(
            presigned_url=url,
            source_file=str(self.package_path),
            name=self.package_path.name,
            total_size=self.package_path.stat().st_size,
        )
        uploader.upload()


def parse_distributions(script_args, args):
    distributions = {}
    for arg in args:
        is_distribution = False
        for p in [UniformDistributionParser, LogUniformDistributionParser, CategoricalDistributionParser]:
            if p.is_distribution(arg.replace("--", "")):
                distributions.update(p.parse(arg.replace("--", "")))
                is_distribution = True
                break
        if not is_distribution:
            script_args.append(arg)

    return {k: Distributions(distribution=x["distribution"], params=x["params"]) for k, x in distributions.items()}


def parse_args(args):
    parsed = {}
    last_arg = None

    # build a dict where key = arg, value = value of the arg or None if just a flag
    new_args = []
    for a in args:
        if "=" in a:
            new_args.extend(a.split("="))
        else:
            new_args.append(a)

    for arg_candidate in new_args:

        # only look at --keys
        if "--" not in arg_candidate:
            parsed[last_arg].append(arg_candidate)
        else:
            parsed[arg_candidate] = []
            last_arg = arg_candidate

    return {k: " ".join(v) for k, v in parsed.items()}


def parse_range_to_categorical(value: str):
    range_args = [literal_eval(val.strip()) for val in value.replace("range(", "").replace(")", "").split(",")]
    # Evaluates range expression to a list.
    value = list(range(*range_args))
    return {"distribution": "categorical", "params": {"choices": value}}


def _parse_list(value: str):
    try:
        value = literal_eval(value)
        if not isinstance(value, list):
            return None

        return value
    except Exception:
        pass


def parse_list_to_categorical(value: str):
    try:
        value = literal_eval(value)
        if not isinstance(value, list):
            return None

        return {"distribution": "categorical", "params": {"choices": value}}
    except Exception:
        pass


def parse_grid_search(script_args, args):
    distributions = {}
    parsed = parse_args(args)

    for key, value in parsed.items():
        expected_value = None
        if RANGE_REGEX.match(str(value)):
            expected_value = parse_range_to_categorical(value)
        else:
            expected_value = parse_list_to_categorical(value)
        if expected_value:
            distributions[key] = expected_value
        else:
            script_args.append(f"{key}={value}")
    return distributions


def parse_random_search(script_args, args):
    distributions = {}
    parsed = parse_args(args)

    for key, value in parsed.items():
        expected_value = None

        if RANGE_REGEX.match(value):
            expected_value = parse_range_to_categorical(value)
        else:
            expected_value = parse_list_to_categorical(value)
            if not expected_value:
                for p in [UniformDistributionParser, LogUniformDistributionParser, CategoricalDistributionParser]:
                    if p.is_distribution(value):
                        try:
                            expected_value = p.parse(f"{key}={value}")[key]
                            break
                        except Exception:
                            pass

        if expected_value:
            distributions[key] = expected_value
        else:
            script_args.append(f"{key}={value}")
    return distributions


def parse_hydra(script_args, args):
    distributions = {}
    parser = OverridesParser.create()
    parsed = parser.parse_overrides(args)
    for override in parsed:
        if override.is_sweep_override():
            key = override.get_key_element()
            values = override.sweep_string_iterator()
            distributions[key] = {"distribution": "categorical", "params": {"choices": list(values)}}
        else:
            script_args.append(override.input_line)
    return distributions


class RunSweepCommand(ClientCommand):

    description = """Run a sweep by providing a script, the cloud compute type and optional
                     data entries to be made available at a given path. Hyperparameters can be
                     provided as lists (`model.lr=\"[0.01, 0.1]\"`) or using distributions
                     (`model.lr=\"uniform(0.01, 0.1)\"`, `model.lr=\"log_uniform(0.01, 0.1)\"`).
                     Hydra multirun override syntax is also supported."""
    requirements = ["traitlets", "lightning_hpo"]
    SUPPORTED_DISTRIBUTIONS = ("uniform", "log_uniform", "categorical")

    def run(self) -> None:
        parser = ArgumentParser()

        parser.add_argument("script_path", type=str, help="The path to the script to run.")
        parser.add_argument(
            "--algorithm",
            default="grid_search",
            type=str,
            help="The search algorithm to use.",
            choices=["grid_search", "random_search", "bayesian"],
        )
        parser.add_argument("--total_experiments", default=None, type=int, help="The total number of experiments")
        parser.add_argument("--parallel_experiments", default=None, type=int, help="Number of experiments to run.")
        parser.add_argument(
            "--requirements",
            nargs="+",
            default=[],
            help="List of requirements separated by a comma or requirements.txt filepath.",
        )
        parser.add_argument(
            "--packages",
            nargs="+",
            default=[],
            help="List of system packages to be installed via apt install, separated by a comma.",
        )
        parser.add_argument(
            "--cloud_compute",
            default="cpu",
            choices=["cpu", "cpu-small", "cpu-medium", "gpu", "gpu-fast", "gpu-fast-multi"],
            type=str,
            help="The machine to use in the cloud.",
        )
        parser.add_argument("--name", default=None, type=str, help="Configure your sweep name.")
        parser.add_argument(
            "--logger",
            default="tensorboard",
            choices=["tensorboard", "wandb"],
            type=str,
            help="The logger to use with your sweep.",
        )
        parser.add_argument(
            "--direction",
            default="minimize",
            choices=["minimize", "maximize"],
            type=str,
            help="In which direction to optimize.",
        )
        parser.add_argument(
            "--framework",
            default="pytorch_lightning",
            choices=["pytorch_lightning", "base"],
            type=str,
            help="Which framework you are using.",
        )
        parser.add_argument(
            "--num_nodes",
            default=1,
            type=int,
            help="The number of nodes.",
        )
        parser.add_argument(
            "--disk_size",
            default=10,
            type=int,
            help="The disk size in Gigabytes.",
        )
        parser.add_argument(
            "--data",
            nargs="+",
            default=[],
            help="Provide a list of Data (and optionally the mount_path in the format `<name>:<mount_path>`) to mount to the experiments.",
        )
        parser.add_argument(
            "--syntax",
            default="default",
            choices=["default", "hydra"],
            type=str,
            help="Syntax for sweep parameters at the CLI.",
        )
        parser.add_argument(
            "--pip-install-source",
            default=False,
            action="store_true",
            help="Run `pip install -e .` on the uploaded source before running",
        )

        hparams, args = parser.parse_known_args()

        if hparams.framework != "pytorch_lightning" and hparams.num_nodes > 1:
            raise ValueError("Multi Node is support only for PyTorch Lightning.")

        script_args = []

        total_experiments = hparams.total_experiments

        if hparams.syntax == "hydra":
            distributions = parse_hydra(script_args, args)
            total_experiments = -1
        elif hparams.algorithm == "grid_search":
            distributions = parse_grid_search(script_args, args)
            total_experiments = -1
        elif hparams.algorithm == "random_search":
            distributions = parse_random_search(script_args, args)
            if not distributions:
                total_experiments = 1
            elif distributions and hparams.total_experiments is None:
                raise Exception("Please, specify the `total_experiments`.")
            else:
                total_experiments = hparams.total_experiments
            if hparams.parallel_experiments is None:
                hparams.parallel_experiments = total_experiments
        else:
            distributions = parse_distributions(script_args, args)

        name = hparams.name or str(uuid4()).split("-")[-1][:7]

        if not os.path.exists(hparams.script_path):
            raise ValueError(f"The provided script doesn't exist: {hparams.script_path}")

        if not os.path.exists(hparams.script_path):
            raise FileNotFoundError(f"The provided script doesn't exist: {hparams.script_path}")

        if len(hparams.requirements) == 1 and os.path.isfile(hparams.requirements[0]):
            with open(hparams.requirements, "r") as f:
                hparams.requirements = [line.replace("\n", "") for line in f.readlines() if line.strip()]

        repo = CustomLocalSourceCodeDir(path=Path(hparams.script_path).parent.resolve())
        repo.package()
        repo.upload(url=f"{self.app_url}/api/v1/upload_file/{name}")

        data_split = [data.split(":") if ":" in data else (data, None) for data in hparams.data]
        data = {data[0]: data[1] for data in data_split}

        config = SweepConfig(
            sweep_id=name,
            script_path=hparams.script_path,
            total_experiments=int(total_experiments),
            parallel_experiments=hparams.parallel_experiments if isinstance(hparams.parallel_experiments, int) else 1,
            requirements=hparams.requirements,
            packages=hparams.packages,
            script_args=script_args,
            distributions=distributions,
            algorithm=hparams.algorithm,
            framework=hparams.framework,
            cloud_compute=hparams.cloud_compute,
            num_nodes=hparams.num_nodes,
            logger=hparams.logger,
            direction=hparams.direction,
            experiments={},
            disk_size=hparams.disk_size,
            pip_install_source=hparams.pip_install_source,
            data=data,
            username=getuser(),
        )
        response = self.invoke_handler(config=config)
        print(response)



from argparse import ArgumentParser

from lightning.app.utilities.commands import ClientCommand
from pydantic import BaseModel


class StopSweepConfig(BaseModel):
    sweep_id: str


class StopSweepCommand(ClientCommand):

    description = "Stop all experiments in a sweep. Note that currently sweeps cannot be resumed."

    def run(self) -> None:
        parser = ArgumentParser()
        parser.add_argument("--name", type=str, help="The associated `sweep_id` to stop.")
        hparams = parser.parse_args()
        response = self.invoke_handler(config=StopSweepConfig(sweep_id=hparams.name))
        print(response)



import pytorch_lightning as pl
from lightning_transformers.task.nlp.language_modeling import LanguageModelingDataModule, LanguageModelingTransformer
from transformers import AutoTokenizer

model_name = "gpt2-xl"

tokenizer = AutoTokenizer.from_pretrained(model_name)

model = LanguageModelingTransformer(
    pretrained_model_name_or_path=model_name,
    tokenizer=tokenizer,
    deepspeed_sharding=True,  # defer initialization of the model to shard/load pre-train weights
)

dm = LanguageModelingDataModule(
    batch_size=1,
    dataset_name="wikitext",
    dataset_config_name="wikitext-2-raw-v1",
    tokenizer=tokenizer,
)
trainer = pl.Trainer(
    accelerator="auto",
    devices="auto",
    strategy="deepspeed_stage_3",
    precision=16,
    max_epochs=20,
)

trainer.fit(model, dm)



from lightning import LightningFlow
from lightning.app.frontend.stream_lit import StreamlitFrontend
from lightning.app.utilities.state import AppState


class HiPlotFlow(LightningFlow):
    def __init__(self):
        super().__init__()
        self.data = []

    def configure_layout(self):
        return StreamlitFrontend(render_fn=render_fn)


def render_fn(state: AppState):
    import json

    import hiplot as hip
    import streamlit as st

    st.set_page_config(layout="wide")

    if not state.data:
        st.write("No data available yet ! Stay tuned")
        return

    xp = hip.Experiment.from_iterable(state.data)
    ret_val = xp.to_streamlit(ret="selected_uids", key="hip").display()
    st.markdown("hiplot returned " + json.dumps(ret_val))



from abc import ABC, abstractmethod
from typing import Any, Dict, List, Optional

from lightning import LightningFlow
from lightning.app.utilities.tracer import Tracer


class Logger(ABC):
    def on_after_experiment_start(self, sweep_id: str):
        ...

    @abstractmethod
    def on_after_experiment_end(
        self, sweep_id: str, experiment_id: int, monitor: Optional[str], score: Optional[float], params: Dict[str, Any]
    ):
        ...

    @abstractmethod
    def connect(self, flow: LightningFlow):
        ...

    @abstractmethod
    def configure_layout(self) -> Dict[str, Any]:
        ...

    @abstractmethod
    def configure_tracer(self, tracer, sweep_id: str, experiment_id: int, experiment_name: str, params: Dict[str, Any]):
        ...

    @abstractmethod
    def get_url(self, experiment_id: int) -> Optional[str]:
        ...


class NoneLogger(Logger):
    def on_after_experiment_start(self, sweep_id: str):
        pass

    def on_after_experiment_end(self, *args, **kwargs):
        pass

    def connect(self, flow: LightningFlow):
        pass

    def configure_layout(self) -> List:
        return []

    def configure_tracer(self, tracer, sweep_id: str, experiment_id: int, params: Dict[str, Any]):
        return Tracer()

    def get_url(self, experiment_id: int) -> Optional[str]:
        pass



import optuna
from lightning import LightningApp
from sklearn import datasets
from sklearn.linear_model import SGDClassifier
from sklearn.model_selection import train_test_split

from lightning_hpo import Objective, Sweep
from lightning_hpo.algorithm.optuna import OptunaAlgorithm
from lightning_hpo.distributions.distributions import LogUniform


class MyObjective(Objective):
    def objective(self, alpha: float):

        iris = datasets.load_iris()
        classes = list(set(iris.target))
        train_x, valid_x, train_y, valid_y = train_test_split(
            iris.data,
            iris.target,
            test_size=0.25,
            random_state=0,
        )

        clf = SGDClassifier(alpha=alpha)

        self.monitor = "accuracy"

        for step in range(100):
            clf.partial_fit(train_x, train_y, classes=classes)
            intermediate_value = clf.score(valid_x, valid_y)

            # WARNING: Assign to reports,
            # so the state is instantly sent to the flow.
            self.reports = self.reports + [[intermediate_value, step]]

        self.best_model_score = clf.score(valid_x, valid_y)


app = LightningApp(
    Sweep(
        objective_cls=MyObjective,
        total_experiments=20,
        parallel_experiments=1,
        algorithm=OptunaAlgorithm(
            optuna.create_study(
                pruner=optuna.pruners.MedianPruner(),
                direction="maximize",
            )
        ),
        distributions={"alpha": LogUniform(1e-5, 1e-1)},
    )
)



import argparse
import os
import re
from typing import Dict, List, Optional

import rich
from lightning.app.storage.path import _filesystem, _shared_storage_path
from lightning.app.utilities.commands import ClientCommand
from pydantic import BaseModel
from rich.color import ANSI_COLOR_NAMES
from rich.text import Text
from rich.tree import Tree


def _walk_folder(tree: Tree, sorted_directories: Dict[str, List[str]], root_path: str, depth: int):
    parents = {}
    for directory in sorted_directories:
        splits = directory.split("/")
        if depth == len(splits):
            for path in sorted_directories[directory]:
                text_filename = Text(path.replace(directory + "/", ""), "green")
                text_filename.highlight_regex(r"\..*$", "bold red")
                text_filename.stylize(f"link {path}")
                icon = "🐍 " if path.endswith(".py") else "📄 "
                tree.add(Text(icon) + text_filename)
        else:
            root_folder = "/".join(splits[1 : depth + 1])  # noqa E203
            if root_folder not in parents:
                parents[root_folder] = []
            parents[root_folder].append(directory)

    if not parents:
        return

    for root_folder, directories in parents.items():
        folder = root_folder.split("/")[-1]
        style = "dim" if root_folder.startswith("__") else ""
        branch = tree.add(
            f"[bold magenta]:open_file_folder: {folder}",
            style=style,
            guide_style=style,
        )
        _walk_folder(
            branch,
            {directory: sorted_directories[directory] for directory in directories},
            os.path.join(root_path, folder),
            depth + 1,
        )


def walk_folder_tree(paths: List[str], tree: Tree) -> None:
    """Recursively build a Tree with directory contents."""
    paths = sorted(paths)
    directories = {}
    for p in paths:
        directory = os.path.dirname(p)
        if directory not in directories:
            directories[directory] = []
        directories[directory].append(p)

    _walk_folder(tree, directories, "artifacts", 1)
    rich.print(tree)


def walk_folder_flat_(paths: List[str]) -> None:
    colors = list(ANSI_COLOR_NAMES)
    paths = sorted(paths)
    mapping = {}
    counter = 1
    showed = {}
    for path in paths:
        splits = path.split("/")

        for idx in range(1, len(splits)):
            directory = "/".join(splits[:idx])
            if directory not in showed:
                showed[directory] = True
                rich.print(directory)

        for split in splits:
            if split not in mapping:
                mapping[split] = colors[counter]
                counter += 1

        colored_splits = []
        for split_idx, split in enumerate(splits, 1):
            color = mapping[split] if split_idx == len(splits) else colors[0]
            colored_split = f"[{color}]{split}[/{color}]"
            colored_splits.append(colored_split)
        colored_path = "/".join(colored_splits)
        rich.print(colored_path)


def _walk_folder_flat(tree: List[str], sorted_directories: Dict[str, List[str]], root_path: str, depth: int):
    parents = {}
    for directory in sorted_directories:
        splits = directory.split("/")
        if depth == len(splits):
            for path in sorted_directories[directory]:
                tree.append(path)
        else:
            root_folder = "/".join(splits[1 : depth + 1])  # noqa E203
            if root_folder not in parents:
                parents[root_folder] = []
            parents[root_folder].append(directory)

    if not parents:
        return

    for root_folder, directories in parents.items():
        folder = root_folder.split("/")[-1]
        root_folder = f":open_file_folder: /{root_folder}/"
        tree.append(f"[black]{root_folder}[/black]")
        _walk_folder_flat(
            tree,
            {directory: sorted_directories[directory] for directory in directories},
            os.path.join(root_path, folder),
            depth + 1,
        )


def _add_colors(filename: str) -> str:
    colors = list(ANSI_COLOR_NAMES)
    color = "magenta"

    if ".yaml" in filename:
        color = colors[1]

    elif ".ckpt" in filename:
        color = colors[2]

    elif "events.out.tfevents" in filename:
        color = colors[3]

    elif ".py" in filename:
        color = colors[4]

    elif ".png" in filename:
        color = colors[5]

    return f"[{color}]{filename}[/{color}]"


def walk_folder_flat(paths: List[str]) -> None:
    """Recursively build a Tree with directory contents."""
    paths = sorted(paths)
    directories = {}
    tree = []
    for p in paths:
        directory = os.path.dirname(p)
        if directory not in directories:
            directories[directory] = []
        directories[directory].append(p)

    _walk_folder_flat(tree, directories, "artifacts", 1)

    for path in tree:
        if not path.startswith("[black"):
            splits = path.split("/")
            path = f'   [black]{"/".join(splits[:-1])}[/black]/{_add_colors(splits[-1])}'
        rich.print(path)


class ShowArtifactsConfig(BaseModel):
    include: Optional[str] = None
    exclude: Optional[str] = None


class ShowArtifactsConfigResponse(BaseModel):
    sweep_names: List[str]
    experiment_names: List[str]
    paths: List[str]


class ShowArtifactsCommand(ClientCommand):

    description = "Show artifacts for experiments or sweeps, in flat or tree layout."

    def run(self) -> None:
        # 1. Parse the user arguments.
        parser = argparse.ArgumentParser()
        parser.add_argument("--names", nargs="+", default=[], help="Provide a list of experiments or sweep names.")
        parser.add_argument(
            "--display_mode",
            default="flat",
            choices=["flat", "tree"],
            help="Structure of the files displayed in the terminal",
        )
        hparams = parser.parse_args()

        config = ShowArtifactsConfig(include=None, exclude=None)
        response = ShowArtifactsConfigResponse(**self.invoke_handler(config=config))

        for name in hparams.names:
            has_found = False
            if name in response.sweep_names:
                has_found = True
            elif name in response.experiment_names:
                has_found = True

            if not has_found:
                raise ValueError(
                    f"The provided name `{name}` doesn't exists. "
                    f" Here are the available Sweeps {response.sweep_names} and "
                    f"available Experiments {response.experiment_names}."
                )

        include = None
        if hparams.names:
            include = "|".join(hparams.names)

        paths = _filter_paths(response.paths, include=include, exclude=None)

        if hparams.display_mode == "tree":
            tree = Tree(
                "[bold magenta]:open_file_folder: root",
                guide_style="bold bright_blue",
            )

            walk_folder_tree(paths, tree)
        else:
            walk_folder_flat(paths)


def _filter_paths(paths: List[str], include: Optional[str], exclude: Optional[str]) -> List[str]:
    out = []

    if include:
        include_pattern = re.compile(include)

    if exclude:
        exclude_pattern = re.compile(exclude)

    for path in paths:
        if include or exclude:
            if exclude and len(exclude_pattern.findall(path)):
                pass

            elif (include and len(include_pattern.findall(path))) or include is None:
                out.append(path)
        else:
            out.append(path)
    return out


def _collect_artifact_paths(config: ShowArtifactsConfig, replace: bool = True) -> List[str]:
    """This function is responsible to collecting the files from the shared filesystem."""
    fs = _filesystem()
    paths = []

    shared_storage = _shared_storage_path()
    for root_dir, _, files in fs.walk(shared_storage):
        if replace:
            root_dir = str(root_dir).replace(str(shared_storage), "").replace("/artifacts/drive", "")
        for f in files:
            paths.append(os.path.join(str(root_dir), f))

    return _filter_paths(paths, config.include, config.exclude)



import urllib.parse
from typing import List, Optional

from lightning_hpo.commands.notebook.run import NotebookConfig, RunNotebookCommand
from lightning_hpo.commands.notebook.show import ShowNotebookCommand
from lightning_hpo.commands.notebook.stop import StopNotebookCommand, StopNotebookConfig
from lightning_hpo.components.notebook import JupyterLab
from lightning_hpo.controllers.controller import Controller
from lightning_hpo.utilities.enum import Stage


class NotebookController(Controller):

    model = NotebookConfig

    def on_reconcile_start(self, configs: List[NotebookConfig]):
        for config in configs:
            work_name = urllib.parse.quote_plus(config.notebook_name)
            if work_name in self.r and self.r[work_name].status.stage == Stage.STOPPED:
                self.r[work_name].stage = Stage.STOPPED

            if config.desired_stage == Stage.RUNNING and config.stage not in [Stage.PENDING, Stage.RUNNING]:
                self.r[work_name] = JupyterLab(
                    kernel="python",
                    config=config,
                )
                self.r[work_name].stage = Stage.PENDING
            elif all(
                [
                    work_name in self.r,
                    config.desired_stage == Stage.STOPPED,
                    config.stage not in [Stage.STOPPING, Stage.STOPPED],
                ]
            ):
                self.r[work_name].stop()
                self.r[work_name]._url = ""
                self.r[work_name].stage = Stage.STOPPING

    def run_notebook(self, config: NotebookConfig) -> str:
        matched_notebook: Optional[NotebookConfig] = None
        for existing_config in self.db.select_all():
            if existing_config.notebook_name == config.notebook_name:
                matched_notebook = existing_config

        if matched_notebook:
            if matched_notebook.stage == Stage.RUNNING and matched_notebook.desired_stage == Stage.RUNNING:
                return f"The notebook `{config.notebook_name}` is already running."

            config.desired_stage = Stage.RUNNING
            # Update config in the database
            self.db.update(config)
            return f"The notebook `{config.notebook_name}` has been updated."

        self.db.insert(config)
        return f"The notebook `{config.notebook_name}` has been created."

    def stop_notebook(self, config: StopNotebookConfig) -> str:
        matched_notebook = None
        for notebook in self.r.values():
            if notebook.notebook_name == config.notebook_name:
                matched_notebook = notebook

        if matched_notebook:
            if matched_notebook.desired_stage != Stage.STOPPED:
                matched_notebook.desired_stage = Stage.STOPPED
                self.db.update(matched_notebook.collect_model())
                return f"The notebook `{config.notebook_name}` has been stopped."
            return f"The notebook `{config.notebook_name}` is already stopped."
        names = [notebook.notebook_name for notebook in self.r.values()]
        if not names:
            return "You don't have any notebooks. Create a notebook with `lightning run notebook --name=my_name`."
        return f"The notebook `{config.notebook_name}` doesn't exist. Found the following notebooks: {names}."

    def show_notebook(self):
        if self.db_url:
            return self.db.select_all()
        return []

    def configure_commands(self) -> List:
        return [
            {"run notebook": RunNotebookCommand(self.run_notebook)},
            {"stop notebook": StopNotebookCommand(self.stop_notebook)},
            {"show notebooks": ShowNotebookCommand(self.show_notebook)},
        ]



import os
from pathlib import Path
from subprocess import Popen
from uuid import uuid4

from lightning import LightningWork
from lightning.app.storage import Drive
from lightning.app.storage.path import _filesystem

from lightning_hpo.commands.tensorboard.stop import TensorboardConfig
from lightning_hpo.controllers.controller import ControllerResource
from lightning_hpo.utilities.enum import Stage


class Tensorboard(LightningWork, ControllerResource):

    model = TensorboardConfig

    def __init__(self, *args, drive: Drive, config: TensorboardConfig, **kwargs):
        super().__init__(*args, parallel=True, **kwargs)
        self.drive = drive
        self.sweep_id = config.sweep_id
        self.shared_folder = config.shared_folder
        self.stage = config.stage
        self.desired_stage = config.desired_stage
        self.config = config.dict()

    def run(self):
        use_localhost = "LIGHTNING_APP_STATE_URL" not in os.environ

        local_folder = f"./tensorboard_logs/{uuid4()}"

        os.makedirs(local_folder, exist_ok=True)

        # Note: Used tensorboard built-in sync methods but it doesn't seem to work.
        cmd = f"tensorboard --logdir={local_folder} --host {self.host} --port {self.port}"
        self._process = Popen(cmd, shell=True, env=os.environ)

        self.stage = Stage.RUNNING
        fs = _filesystem()
        root_folder = str(self.drive.drive_root)

        while True:
            fs.invalidate_cache()
            for dir, _, files in fs.walk(root_folder):
                for filepath in files:
                    if "events.out.tfevents" not in filepath:
                        continue
                    source_path = os.path.join(dir, filepath)
                    target_path = os.path.join(dir, filepath).replace(root_folder, local_folder)
                    if use_localhost:
                        parent = Path(target_path).resolve().parent
                        if not parent.exists():
                            parent.mkdir(exist_ok=True, parents=True)
                    fs.get(source_path, str(Path(target_path).resolve()))

    def on_exit(self):
        assert self._process
        self._process.kill()

    def on_collect_model(self, model_dict):
        model_dict["url"] = self.url



from typing import List

from lightning.app.utilities.commands import ClientCommand
from rich.console import Console
from rich.table import Table

from lightning_hpo.commands.notebook.run import NotebookConfig


def _show_notebooks(notebooks: List[NotebookConfig]):
    table = Table(
        "name",
        "status",
        "cloud_compute",
        "requirements",
        title="Notebooks",
        show_header=True,
        header_style="bold green",
    )

    for notebook in notebooks:
        table.add_row(
            notebook.notebook_name,
            notebook.stage,
            notebook.cloud_compute,
            str(notebook.requirements),
        )
    console = Console()
    console.print(table)


class ShowNotebookCommand(ClientCommand):
    def run(self) -> None:
        response = self.invoke_handler()
        _show_notebooks([NotebookConfig(**notebook) for notebook in response])



import os
from argparse import ArgumentParser
from getpass import getuser
from pathlib import Path
from uuid import uuid4

from lightning.app.utilities.commands import ClientCommand

from lightning_hpo.commands.sweep.run import CustomLocalSourceCodeDir, ExperimentConfig, SweepConfig


class RunExperimentCommand(ClientCommand):

    description = """Run an experiment by providing a script, the cloud compute type and optional
                     data entries to be made available at a given path."""

    def run(self) -> None:
        parser = ArgumentParser()

        parser.add_argument("script_path", type=str, help="The path to the script to run.")
        parser.add_argument(
            "--requirements",
            nargs="+",
            default=[],
            help="List of requirements separated by a comma or requirements.txt filepath.",
        )
        parser.add_argument(
            "--packages",
            nargs="+",
            default=[],
            help="List of system packages to be installed via apt install, separated by a comma.",
        )
        parser.add_argument(
            "--cloud_compute",
            default="cpu",
            choices=["cpu", "cpu-small", "cpu-medium", "gpu", "gpu-fast", "gpu-fast-multi"],
            type=str,
            help="The machine to use in the cloud.",
        )
        parser.add_argument("--name", default=None, type=str, help="The name of the experiment.")
        parser.add_argument(
            "--logger",
            default="tensorboard",
            choices=["tensorboard", "wandb"],
            type=str,
            help="The logger to use with your sweep.",
        )
        parser.add_argument(
            "--num_nodes",
            default=1,
            type=int,
            help="The number of nodes.",
        )
        parser.add_argument(
            "--disk_size",
            default=80,
            type=int,
            help="The disk size in Gigabytes.",
        )
        parser.add_argument(
            "--data",
            nargs="+",
            default=[],
            help="Provide a list of Data (and optionally the mount_path in the format `<name>:<mount_path>`) to mount to the experiment.",
        )
        parser.add_argument(
            "--framework",
            default="pytorch_lightning",
            choices=["pytorch_lightning", "base"],
            type=str,
            help="Which framework you are using.",
        )
        parser.add_argument(
            "--pip-install-source",
            default=False,
            action="store_true",
            help="Run `pip install -e .` on the uploaded source before running",
        )

        hparams, args = parser.parse_known_args()

        name = hparams.name or str(uuid4()).split("-")[-1][:7]

        if not os.path.exists(hparams.script_path):
            raise FileNotFoundError(f"The provided script doesn't exist: {hparams.script_path}")

        if len(hparams.requirements) == 1 and Path(hparams.requirements[0]).resolve().exists():
            requirements_path = Path(hparams.requirements[0]).resolve()
            with open(requirements_path, "r") as f:
                hparams.requirements = [line.replace("\n", "") for line in f.readlines() if line.strip()]

        repo = CustomLocalSourceCodeDir(path=Path(hparams.script_path).parent.resolve())
        repo.package()
        repo.upload(url=f"{self.app_url}/api/v1/upload_file/{name}")

        data_split = [data.split(":") if ":" in data else (data, None) for data in hparams.data]
        data = {data[0]: data[1] for data in data_split}

        config = SweepConfig(
            sweep_id=name,
            script_path=hparams.script_path,
            total_experiments=1,
            parallel_experiments=1,
            requirements=hparams.requirements,
            packages=hparams.packages,
            script_args=args,
            distributions={},
            algorithm="",
            framework=hparams.framework,
            cloud_compute=hparams.cloud_compute,
            num_nodes=hparams.num_nodes,
            logger=hparams.logger,
            direction="minimize",  # This won't be used
            experiments={0: ExperimentConfig(name=name, params={})},
            disk_size=hparams.disk_size,
            pip_install_source=hparams.pip_install_source,
            data=data,
            username=str(getuser()),
        )
        response = self.invoke_handler(config=config)
        print(response)



from lightning_hpo.framework.agnostic import Objective
from lightning_hpo.framework.pytorch_lightning import ObjectiveLightningTrainingComponent

_OBJECTIVE_FRAMEWORK = {"base": Objective, "pytorch_lightning": ObjectiveLightningTrainingComponent}



import logging
from typing import Any, Dict, List, Optional, Tuple

import optuna
from lightning_utilities.core.apply_func import apply_to_collection
from optuna import create_study, Study, Trial
from optuna.distributions import (
    BaseDistribution,
    CategoricalDistribution,
    IntUniformDistribution,
    LogUniformDistribution,
    UniformDistribution,
)

from lightning_hpo.algorithm.base import Algorithm
from lightning_hpo.distributions import DistributionDict
from lightning_hpo.distributions.distributions import Distribution

_logger = logging.getLogger(__name__)

_DISTRIBUTION_TO_OPTUNA = {
    "uniform": UniformDistribution,
    "int_uniform": IntUniformDistribution,
    "log_uniform": LogUniformDistribution,
    "categorical": CategoricalDistribution,
}


class OptunaAlgorithm(Algorithm):
    def __init__(self, study: Optional[Study] = None, direction: Optional[str] = "minimize") -> None:
        self.study = study or create_study(direction=direction)
        self.experiments: Dict[int, Trial] = {}
        self.reports = {}
        self.distributions: Dict[str, BaseDistribution] = {}

    def register_distributions(self, distributions: Dict[str, DistributionDict]):
        for var_name, distribution in distributions.items():
            distribution_cls = _DISTRIBUTION_TO_OPTUNA[distribution["distribution"]]
            distribution = distribution_cls(**distribution["params"])
            self.distributions[var_name] = distribution

    def register_experiments(self, experiments_config: List[Dict]) -> None:
        for experiment_config in experiments_config:
            trial = optuna.trial.create_trial(
                params=experiment_config["params"],
                distributions=self.distributions,
                value=experiment_config["best_model_score"],
            )
            self.study.add_trial(trial)

    def experiment_start(self, experiment_id: int) -> None:
        if experiment_id not in self.experiments:
            self.experiments[experiment_id] = self.study.ask(self.distributions)

    def experiment_end(self, experiment_id: int, score: float):
        try:
            self.study.tell(experiment_id, score)
        except RuntimeError as e:
            # The trial has already been added to the study.
            print(e)
            pass

        _logger.info(
            f"Experiment {experiment_id} finished with value: {score} and parameters: {self.experiments[experiment_id].params}. "  # noqa: E501
            f"Best is experiment {self.study.best_trial.number} with value: {self.study.best_trial.value}."
        )

    def should_prune(self, experiment_id: int, reports: List[Tuple[float, int]]) -> bool:
        trial = self.experiments[experiment_id]
        if experiment_id not in self.reports:
            self.reports[experiment_id] = []

        for report in reports:
            if report in self.reports[experiment_id]:
                continue
            trial.report(*report)
            self.reports[experiment_id].append(report)

            if trial.should_prune():
                _logger.info(f"Trial {experiment_id} pruned.")
                return True

        return False

    def get_params(self, experiment_id: int) -> Dict[str, Any]:
        params = self.experiments[experiment_id].params
        out = {}
        for k, v in params.items():
            if isinstance(v, float) and v == int(v):
                out[k] = int(v)
            else:
                out[k] = v
        return out


class GridSearch(Algorithm):
    def __init__(self, search_space: Dict[str, Any]):
        sampler = optuna.samplers.GridSampler(search_space)
        self.experiments = {
            i: {n: v for n, v in zip(sampler._param_names, sampler._all_grids[i])}
            for i in range(len(sampler._all_grids))
        }

    @property
    def total_experiments(self) -> int:
        return len(self.experiments)

    def experiment_start(self, experiment_id: int) -> None:
        pass

    def register_experiments(self, experiments_config: List[Dict]) -> None:
        pass

    def register_distributions(self, distributions):
        pass

    def get_params(self, experiment_id: int) -> Dict[str, Any]:
        return self.experiments[experiment_id]

    def should_prune(self) -> bool:
        return False

    def experiment_end(self, experiment_id: int, score: float):
        pass


class RandomSearch(Algorithm):
    def __init__(self, distributions: Dict[str, Any]):
        self.study = create_study(sampler=optuna.samplers.RandomSampler())
        self.distributions = {}
        distributions = apply_to_collection(distributions, Distribution, lambda x: x.to_dict())
        self._register_distributions(distributions)
        self.experiments = {}

    def _register_distributions(self, distributions: Dict[str, DistributionDict]):
        for var_name, distribution in distributions.items():
            distribution_cls = _DISTRIBUTION_TO_OPTUNA[distribution["distribution"]]
            distribution = distribution_cls(**distribution["params"])
            self.distributions[var_name] = distribution

    def experiment_start(self, experiment_id: int) -> None:
        self.experiments[experiment_id] = self.study.ask(self.distributions)

    def register_experiments(self, experiments_config: List[Dict]) -> None:
        for experiment_config in experiments_config:
            trial = optuna.trial.create_trial(
                params=experiment_config["params"],
                distributions=self.distributions,
                value=experiment_config["best_model_score"],
            )
            self.study.add_trial(trial)

    def register_distributions(self, distributions):
        assert not distributions

    def get_params(self, experiment_id: int) -> Dict[str, Any]:
        params = self.experiments[experiment_id].params
        out = {}
        for k, v in params.items():
            if isinstance(v, float) and v == int(v):
                out[k] = int(v)
            else:
                out[k] = v
        return out

    def should_prune(self) -> bool:
        return False

    def experiment_end(self, experiment_id: int, score: float):
        pass



import os
import subprocess
import sys
import time
import uuid
from typing import Any, Dict, Optional

import pytorch_lightning
from lightning.app.components.training import LightningTrainerScript, PyTorchLightningScriptRunner
from lightning.app.storage import Path

from lightning_hpo.framework.agnostic import Objective


class PyTorchLightningObjective(Objective, PyTorchLightningScriptRunner):

    """This component executes a PyTorch Lightning script
    and injects a callback in the Trainer at runtime in order to start tensorboard server."""

    def __init__(
        self,
        *args,
        logger: str,
        sweep_id: str,
        experiment_id: int,
        experiment_name: str,
        num_nodes: int,
        last_model_path: Optional[str] = None,
        pip_install_source: bool = False,
        **kwargs,
    ):
        Objective.__init__(
            self,
            logger=logger,
            sweep_id=sweep_id,
            experiment_id=experiment_id,
            experiment_name=experiment_name,
            **kwargs,
        )
        PyTorchLightningScriptRunner.__init__(self, *args, num_nodes=num_nodes, **kwargs)
        self.progress = None
        self.total_parameters = None
        self.start_time = None
        self.end_time = None
        self.last_model_path = Path(last_model_path) if last_model_path else None
        self.pip_install_source = pip_install_source
        self._rootwd = os.getcwd()

    def configure_tracer(self):
        tracer = Objective.configure_tracer(self)
        if self.node_rank == 0:
            return self.add_metadata_tracker(tracer)
        return tracer

    def run(
        self,
        params: Optional[Dict[str, Any]] = None,
        restart_count: int = 0,
        **kwargs,
    ):
        code_dir = "."
        if self.pip_install_source:
            os.chdir(self._rootwd)
            uid = uuid.uuid4().hex[:8]
            code_dir = f"code-{uid}"
            os.makedirs(code_dir)

        if self.last_model_path:
            self.last_model_path.get(overwrite=True)
        self.params = params
        return PyTorchLightningScriptRunner.run(self, params=params, code_dir=code_dir, **kwargs)

    def on_before_run(self):
        if self.pip_install_source:
            subprocess.check_call([sys.executable, "-m", "pip", "install", "-e", "."])

    def on_after_run(self, script_globals):
        self.end_time = time.time()
        self._on_after_run(script_globals)
        self.best_model_path = str(self.best_model_path)

    def _on_after_run(self, script_globals):
        for v in script_globals.values():
            if isinstance(v, (pytorch_lightning.cli.LightningCLI)):
                trainer = v.trainer
                break
            elif isinstance(v, (pytorch_lightning.Trainer)):
                trainer = v
                break
        else:
            raise RuntimeError("No trainer instance found.")

        self.monitor = trainer.checkpoint_callback.monitor

        if trainer.checkpoint_callback.best_model_score:
            self.best_model_path = Path(trainer.checkpoint_callback.best_model_path)
            self.best_model_score = float(trainer.checkpoint_callback.best_model_score)
        else:
            self.best_model_path = Path(trainer.checkpoint_callback.last_model_path)

        self.has_finished = True

    @classmethod
    def distributions(cls):
        return None

    def add_metadata_tracker(self, tracer):
        from lightning_hpo.utilities.imports import _IS_PYTORCH_LIGHTNING_AVAILABLE

        if _IS_PYTORCH_LIGHTNING_AVAILABLE:
            import pytorch_lightning as pl
            from pytorch_lightning.callbacks import Callback, DeviceStatsMonitor
            from pytorch_lightning.strategies.deepspeed import DeepSpeedStrategy
            from pytorch_lightning.utilities import rank_zero_only
            from pytorch_lightning.utilities.model_summary.model_summary import (
                _is_lazy_weight_tensor,
                get_human_readable_count,
            )
            from pytorch_lightning.utilities.model_summary.model_summary_deepspeed import deepspeed_param_size
        else:
            import lightning.pytorch as pl
            from lightning.pytorch.callbacks import Callback, DeviceStatsMonitor
            from lightning.pytorch.strategies.deepspeed import DeepSpeedStrategy
            from lightning.pytorch.utilities import rank_zero_only
            from lightning.pytorch.utilities.model_summary.model_summary import (
                _is_lazy_weight_tensor,
                get_human_readable_count,
            )
            from lightning.pytorch.utilities.model_summary.model_summary_deepspeed import deepspeed_param_size

        class ProgressCallback(Callback):
            def __init__(self, work):
                self.work = work
                self.work.start_time = time.time()
                self.device_stats_callback = DeviceStatsMonitor(cpu_stats=True)

            def setup(
                self,
                trainer,
                pl_module,
                stage: Optional[str] = None,
            ) -> None:
                self.device_stats_callback.setup(trainer, pl_module, stage)
                trainer.checkpoint_callback.save_last = True

            @rank_zero_only
            def on_train_batch_end(self, trainer, pl_module, *args) -> None:
                progress = 100 * (trainer.fit_loop.total_batch_idx + 1) / float(trainer.estimated_stepping_batches)
                if progress > 100:
                    self.work.progress = 100
                else:
                    self.work.progress = round(progress, 4)

                if not self.work.total_parameters:
                    if isinstance(trainer.strategy, DeepSpeedStrategy) and trainer.strategy.zero_stage_3:
                        total_parameters = sum(
                            deepspeed_param_size(p) if not _is_lazy_weight_tensor(p) else 0
                            for p in pl_module.parameters()
                            if p.requires_grad
                        )
                    else:
                        total_parameters = sum(p.numel() for p in pl_module.parameters() if p.requires_grad)
                    human_readable = get_human_readable_count(total_parameters)
                    self.work.total_parameters = str(human_readable)

                if trainer.checkpoint_callback.last_model_path:
                    self.work.last_model_path = Path(trainer.checkpoint_callback.last_model_path)

                self.device_stats_callback.on_train_batch_end(trainer, pl_module, *args)

        def trainer_pre_fn(trainer, *args, **kwargs):
            callbacks = kwargs.get("callbacks", [])
            callbacks.append(ProgressCallback(self))
            kwargs["callbacks"] = callbacks
            return {}, args, kwargs

        def fit_pre_fn(trainer, *args, **kwargs):
            ckpt_path = kwargs.get("ckpt_path", None)
            if not ckpt_path and isinstance(self.last_model_path, Path):
                ckpt_path = str(self.last_model_path)
                print(f"Restarting from checkpoint: {ckpt_path}")
                kwargs["ckpt_path"] = ckpt_path
            return {}, args, kwargs

        tracer.add_traced(pl.Trainer, "__init__", pre_fn=trainer_pre_fn)
        tracer.add_traced(pl.Trainer, "fit", pre_fn=fit_pre_fn)

        return tracer


class ObjectiveLightningTrainingComponent(LightningTrainerScript):
    def __init__(
        self,
        *args,
        experiment_id: int,
        experiment_name: str,
        logger: str,
        sweep_id: str,
        num_nodes: int = 1,
        pip_install_source: bool = False,
        **kwargs,
    ):
        super().__init__(
            *args,
            script_runner=PyTorchLightningObjective,
            logger=logger,
            sweep_id=sweep_id,
            experiment_id=experiment_id,
            experiment_name=experiment_name,
            num_nodes=num_nodes,
            pip_install_source=pip_install_source,
            **kwargs,
        )
        self.experiment_id = experiment_id
        self.experiment_name = experiment_name
        self.has_stopped = False
        self.pruned = False
        self.params = None
        self.restart_count = 0
        self.sweep_id = sweep_id
        self.reports = []
        self.has_stored = False

    def run(
        self, params: Optional[Dict[str, Any]] = None, restart_count: int = 0, last_model_path: Optional[str] = None
    ):
        self.params = params
        self.restart_count = restart_count
        super().run(params=params, restart_count=restart_count, last_model_path=last_model_path)

    @property
    def start_time(self):
        return self.ws[0].start_time

    @property
    def end_time(self):
        return self.ws[0].end_time

    @property
    def total_parameters(self):
        return self.ws[0].total_parameters

    @property
    def progress(self):
        return self.ws[0].progress

    @property
    def monitor(self):
        return self.ws[0].monitor

    @property
    def best_model_path(self):
        return self.ws[0].best_model_path

    @property
    def last_model_path(self):
        return self.ws[0].last_model_path

    @property
    def best_model_score(self):
        return self.ws[0].best_model_score

    @property
    def has_failed(self) -> bool:
        return any(w.has_failed for w in self.works())

    @property
    def status(self):
        return self.ws[0].status

    def stop(self):
        for w in self.works():
            w.stop()
        self.has_stopped = True

    @classmethod
    def distributions(cls):
        return {}



#!/usr/bin/env python

import os
from importlib.util import module_from_spec, spec_from_file_location

from setuptools import find_packages, setup

_PATH_ROOT = os.path.dirname(__file__)


def _load_py_module(fname, pkg="lightning_hpo"):
    spec = spec_from_file_location(os.path.join(pkg, fname), os.path.join(_PATH_ROOT, pkg, fname))
    py = module_from_spec(spec)
    spec.loader.exec_module(py)
    return py


about = _load_py_module("__about__.py")
setup_tools = _load_py_module("setup_tools.py")
long_description = setup_tools._load_readme_description(_PATH_ROOT, homepage=about.__homepage__, ver=about.__version__)

# https://packaging.python.org/discussions/install-requires-vs-requirements /
# keep the meta-data here for simplicity in reading this file... it's not obvious
# what happens and to non-engineers they won't know to look in init ...
# the goal of the project is simplicity for researchers, don't want to add too much
# engineer specific practices

setup(
    name="lightning_hpo",
    version=about.__version__,
    description=about.__docs__,
    author=about.__author__,
    author_email=about.__author_email__,
    url=about.__homepage__,
    download_url="https://github.com/Lightning-AI/lightning-hpo",
    license=about.__license__,
    packages=find_packages(exclude=["tests", "docs"]),
    long_description=long_description,
    long_description_content_type="text/markdown",
    include_package_data=True,
    zip_safe=False,
    keywords=["deep learning", "pytorch", "AI"],
    python_requires=">=3.6",
    entry_points={},
    setup_requires=["wheel"],
    install_requires=setup_tools._load_requirements(_PATH_ROOT),
)



import os
import subprocess
import sys
import uuid
from abc import ABC
from typing import Any, Dict, Optional, TypedDict

from lightning.app.components.python import TracerPythonScript
from lightning.app.utilities.app_helpers import is_overridden

from lightning_hpo.loggers import LoggerType


class ObjectiveResult(TypedDict):
    monitor: str
    score: float
    checkpoint: Optional[str]


class Objective(TracerPythonScript, ABC):
    def objective(self, *args, **kwargs):
        """Override with your own objective logic"""

    def __init__(
        self,
        *args,
        logger: str,
        sweep_id: str,
        experiment_id,
        experiment_name: str,
        raise_exception: bool = False,
        function_name: str = "objective",
        num_nodes: int = 1,  # TODO # Add support for multi node
        last_model_path: Optional[str] = None,
        pip_install_source: bool = False,
        **kwargs,
    ):
        super().__init__(*args, raise_exception=raise_exception, **kwargs)
        self.experiment_id = experiment_id
        self.experiment_name = experiment_name
        self.best_model_score = None
        self.best_model_path = None
        self.params = None
        self.has_told_study = False
        self.reports = []
        self.flow_reports = []
        self.pruned = False
        self.logger = logger
        self._url = None
        self.sweep_id = sweep_id
        self.monitor = None
        self.function_name = function_name
        self.has_stored = False
        self.num_nodes = num_nodes
        self.progress = None
        self.last_model_path = last_model_path
        self.pip_install_source = pip_install_source
        self._rootwd = os.getcwd()

    def configure_tracer(self):
        assert self.params is not None
        tracer = super().configure_tracer()
        LoggerType(self.logger).get_logger().configure_tracer(
            tracer,
            params=self.params,
            sweep_id=self.sweep_id,
            experiment_id=self.experiment_id,
            experiment_name=self.experiment_name,
        )
        return tracer

    def run(self, params: Optional[Dict[str, Any]] = None, restart_count: int = 0):
        if self.pip_install_source:
            os.chdir(self._rootwd)
            uid = uuid.uuid4().hex[:8]
            dirname = f"uploaded-{uid}"
            os.makedirs(dirname)
            os.chdir(dirname)
        self.params = params or {}
        if is_overridden("objective", self, Objective):
            self.objective(**self.params)
        else:
            return super().run(params=params)

    def on_before_run(self):
        if self.pip_install_source:
            subprocess.check_call([sys.executable, "-m", "pip", "install", "-e", "."])

    def on_after_run(self, global_scripts: Any):
        objective_fn = global_scripts.get(self.function_name, None)
        if objective_fn:
            res = objective_fn(**self.params)
            if isinstance(res, (int, float)):
                self.monitor = "score"
                self.best_model_score = round(res, 4)
            elif isinstance(res, dict):
                assert isinstance(res, ObjectiveResult)
        super().on_after_run(global_scripts)



import argparse
import sys
from typing import List

from lightning.app.cli.commands.connection import _retrieve_connection_to_an_app
from lightning.app.cli.commands.logs import _show_logs
from lightning.app.utilities.commands import ClientCommand
from pydantic import BaseModel


class ShowLogsConfig(BaseModel):
    name: str
    components: List[str]


class ShowLogsCommand(ClientCommand):

    description: str = "Show logs of an experiment or a sweep. Optionally follow logs as they stream."

    def run(self):
        parser = argparse.ArgumentParser()
        parser.add_argument(
            "--names",
            nargs="+",
            default=[],
            help="The name or names of the Experiment(s), Sweep(s) or components to show logs from.",
        )
        parser.add_argument("-f", "--follow", action="store_true", default=False)
        hparams = parser.parse_args()

        logs_config: List[ShowLogsConfig] = self.invoke_handler()

        components = []
        sweeps = [c["name"] for c in logs_config if len(c["components"]) > 1]
        experiments = [c["name"] for c in logs_config if len(c["components"]) == 1]

        for config in logs_config:
            for name in hparams.names:
                if name == config["name"]:
                    components.extend(config["components"])

        if not components:
            print(
                f"The provided name {hparams.name} wasn't found."
                f" Here are the sweeps {sweeps} and experiments {experiments}."
            )
            sys.exit(0)

        app_name, _ = _retrieve_connection_to_an_app()

        if app_name == "localhost":
            print("The command `show logs` is currently supported only in the cloud.")
            sys.exit(0)

        _show_logs(app_name, list(set(components)), hparams.follow)



import os
from typing import Any, Dict, Optional

from lightning import LightningFlow

from lightning_hpo.loggers.logger import Logger
from lightning_hpo.utilities.imports import _IS_PYTORCH_LIGHTNING_AVAILABLE, _IS_WANDB_AVAILABLE

if _IS_PYTORCH_LIGHTNING_AVAILABLE:
    import pytorch_lightning
else:
    import lightning.pytorch as pytorch_lightning

if _IS_WANDB_AVAILABLE:
    import wandb


class WandbLogger(Logger):
    def __init__(self):
        super().__init__()
        self._validate_auth()
        os.environ["WANDB_REQUIRE_REPORT_EDITING_V0"] = "1"
        self._api = wandb.Api(api_key=os.environ.get("WANDB_API_KEY"))
        self.entity = os.getenv("WANDB_ENTITY")
        self.sweep_id: Optional[str] = None
        self.storage_id: Optional[str] = None
        self.report_url: Optional[str] = None
        self.report = None

    def on_after_experiment_start(
        self,
        sweep_id: str,
        title: Optional[str] = None,
        desc: Optional[str] = None,
    ):

        from wandb.apis import reports  # noqa F401

        if self.sweep_id:
            return

        self.sweep_id = sweep_id
        wandb.require("report-editing:v0")
        self.report = self._api.create_report(project=self.sweep_id)
        self.report.title = f"{self.sweep_id.title()} Report" if not title else title
        if desc:
            self.report.description = desc

        self.report.save()
        self.storage_id = self.report.id
        self.report_url = f"https://wandb.ai/{self.entity}/{self.sweep_id}/reports/{self.sweep_id}--{self.report.id}"

    def on_after_experiment_end(
        self, sweep_id: str, experiment_id: int, monitor: str, score: float, params: Dict[str, Any]
    ):
        if getattr(self.report, "blocks"):
            return

        panel_grid = wandb.apis.reports.PanelGrid()
        run_set = wandb.apis.reports.RunSet()
        run_set.entity = self.entity
        run_set.project = self.sweep_id
        panel_grid.runsets = [run_set]
        keys = list(params.keys()) + [monitor]
        coords = wandb.apis.reports.ParallelCoordinatesPlot([wandb.apis.reports.PCColumn(p) for p in keys])
        coords.layout = {"x": 0, "y": 0, "w": 24, "h": 10}
        coords.entity = self.entity
        coords.project = self.sweep_id
        panel_grid.panels = [coords]
        run_set.set_filters_with_python_expr(f'User == "{self.entity}"')
        self.report.blocks = [panel_grid]
        self.report.save()

    def connect(self, flow: LightningFlow):
        pass

    def configure_layout(self):
        reports = report = ""
        if self.storage_id is not None:
            reports = f"https://wandb.ai/{self.entity}/{self.sweep_id}/reports/{self.sweep_id}"
            report = f"https://wandb.ai/{self.entity}/{self.sweep_id}/reports/{self.sweep_id}--{self.storage_id}"  # noqa: E501
        return [{"name": "Project", "content": reports}, {"name": "Report", "content": report}]

    def configure_tracer(self, tracer, sweep_id: str, experiment_id: int, experiment_name: str, params: Dict[str, Any]):
        wandb.init(
            project=sweep_id,
            entity=self.entity,
            name=f"experiment_{experiment_id}",
            config=params,
        )

        for k, v in params.items():
            wandb.summary[k] = v

        def trainer_pre_fn(trainer, *args, **kwargs):
            logger = pytorch_lightning.loggers.WandbLogger(
                save_dir=os.path.join(os.getcwd(), "wandb/lightning_logs"),
                project=sweep_id,
                entity=self.entity,
                name=f"experiment_{experiment_id}",
            )
            kwargs["logger"] = [logger]
            return {}, args, kwargs

        tracer.add_traced(pytorch_lightning.Trainer, "__init__", pre_fn=trainer_pre_fn)

    def get_url(self, experiment_id: int) -> None:
        if self.storage_id is not None:
            return f"https://wandb.ai/{self.entity}/{self.sweep_id}/reports/{self.sweep_id}--{self.storage_id}"
        return f"https://wandb.ai/{self.entity}/{self.sweep_id}/reports/{self.sweep_id}"

    @staticmethod
    def _validate_auth():
        if os.getenv("WANDB_API_KEY") is None or os.getenv("WANDB_ENTITY") is None:
            raise Exception(
                "You are trying to use wandb without setting your API key or entity. "
                "HINT: lightning run app app.py --env WANDB_API_KEY=YOUR_API_KEY --env WANDB_ENTITY=YOUR_ENTITY",
            )



import time
from subprocess import Popen
from typing import Optional

from lai_jupyter import JupyterLab
from lightning import CloudCompute

from lightning_hpo.commands.notebook.run import NotebookConfig
from lightning_hpo.controllers.controller import ControllerResource
from lightning_hpo.utilities.enum import Stage


class JupyterLab(JupyterLab, ControllerResource):

    model = NotebookConfig

    def __init__(self, config: NotebookConfig, **kwargs):
        super().__init__(cloud_compute=CloudCompute(name=config.cloud_compute), **kwargs)
        self._process: Optional[Popen] = None

        reqs = self.cloud_build_config.requirements
        self.cloud_build_config.requirements = (reqs if reqs else []) + config.requirements

        self.notebook_name = config.notebook_name
        self.requirements = config.requirements
        self.desired_stage = config.desired_stage
        self.stage = config.stage
        self.start_time = config.start_time

    def run(self, *args, **kwargs):
        super().run()
        self.stage = Stage.RUNNING
        self.start_time = time.time()

    def on_exit(self):
        assert self._process
        self._process.kill()

    def on_collect_model(self, model_dict):
        model_dict["cloud_compute"] = self.cloud_compute.name
        if self.stage == Stage.RUNNING and self.url:
            model_dict["url"] = self.url
        else:
            model_dict.pop("url", None)



import os
import sys

import torch
import torch.nn as nn
import torch.nn.functional as F
import torchvision.transforms as T
from pytorch_lightning import LightningDataModule, LightningModule
from pytorch_lightning.callbacks import ModelCheckpoint
from pytorch_lightning.utilities.cli import LightningCLI
from torchvision.datasets import MNIST


class Net(nn.Module):
    def __init__(self):
        super().__init__()
        self.conv1 = nn.Conv2d(1, 32, 3, 1)
        self.conv2 = nn.Conv2d(32, 64, 3, 1)
        self.dropout1 = nn.Dropout(0.25)
        self.dropout2 = nn.Dropout(0.5)
        self.fc1 = nn.Linear(9216, 128)
        self.fc2 = nn.Linear(128, 10)

    def forward(self, x):
        x = self.conv1(x)
        x = F.relu(x)
        x = self.conv2(x)
        x = F.relu(x)
        x = F.max_pool2d(x, 2)
        x = self.dropout1(x)
        x = torch.flatten(x, 1)
        x = self.fc1(x)
        x = F.relu(x)
        x = self.dropout2(x)
        x = self.fc2(x)
        output = F.log_softmax(x, dim=1)
        return output


class ImageClassifier(LightningModule):
    def __init__(self, model=None, lr=1.0, gamma=0.7):
        super().__init__()
        self.save_hyperparameters(ignore="model")
        print(self.hparams)
        self.model = model or Net()

        checkpoint_path = os.path.join(os.path.dirname(__file__), "demo_weights")
        if os.path.exists(checkpoint_path):
            self.load_state_dict(torch.load(checkpoint_path).state_dict())

    @property
    def example_input_array(self):
        return torch.zeros((1, 1, 28, 28))

    def forward(self, x):
        return self.model(x)

    def training_step(self, batch, batch_idx):
        x, y = batch
        logits = self.forward(x)
        loss = F.nll_loss(logits, y.long())
        self.log("train_loss", loss, on_step=True, on_epoch=True)
        return loss

    def validation_step(self, batch, batch_idx):
        x, y = batch
        logits = self.forward(x)
        loss = F.nll_loss(logits, y.long())
        self.log("val_loss", loss)

    def configure_optimizers(self):
        return torch.optim.Adam(self.model.parameters(), lr=self.hparams.lr)


class MNISTDataModule(LightningDataModule):
    def __init__(self, batch_size: float = 32.0):
        super().__init__()
        self.save_hyperparameters()

    @property
    def transform(self):
        return T.Compose([T.ToTensor(), T.Normalize((0.1307,), (0.3081,))])

    def prepare_data(self) -> None:
        should_download = True
        exists = os.path.exists("./data/MNIST/raw")
        if exists:
            should_download = len(os.listdir("./data/MNIST/raw")) != 8
        if should_download:
            MNIST("./data", download=True)

    def train_dataloader(self):
        train_dataset = MNIST("./data", train=True, download=False, transform=self.transform)
        return torch.utils.data.DataLoader(train_dataset, batch_size=int(self.hparams.batch_size))

    def val_dataloader(self):
        val_dataset = MNIST("./data", train=False, download=False, transform=self.transform)
        return torch.utils.data.DataLoader(val_dataset, batch_size=int(self.hparams.batch_size))


if __name__ == "__main__":
    print(f"The provided arguments are {sys.argv[1:]}")
    cli = LightningCLI(
        ImageClassifier,
        MNISTDataModule,
        seed_everything_default=42,
        save_config_overwrite=True,
        run=False,
        trainer_defaults={
            "max_epochs": 10,
            "limit_train_batches": 20,
            "limit_val_batches": 5,
            "callbacks": [ModelCheckpoint(monitor="val_loss")],
        },
    )
    cli.trainer.fit(cli.model, datamodule=cli.datamodule)
